import bs
import bsUtils
import bsTeamGame
import bsFlag
import bsMap
import bsScoreBoard
import weakref
import bsSpaz
import random
import bsBomb
import bsVector

def getGames():
    return [RaceGame]

class RaceRegion(bs.Actor):
    def __init__(self,activity,pt,index):
        bs.Actor.__init__(self,activity,bs.Actor.UNRETAINED)
        self._pt = pt
        self._index = index
        self.node = bs.newNode("region",
                               owner=self,
                               attrs={'translate':pt[:3],
                                      'scale': (pt[3]*2.0,pt[4]*2.0,pt[5]*2.0),
                                      'type': "box",
                                      'materials':(activity._raceRegionMaterial.materialID,)},
                               delegate=self,
                               name='Race Region')

class RaceGame(bsTeamGame.TeamGameActivity):

    @classmethod
    def getName(c):
        return 'Race'

    @classmethod
    def getDescription(c,sessionType):
        return 'Run real fast!'

    @classmethod
    def getSupportedSessionTypes(self):
        return ['freeForAll','teams']

    @classmethod
    def getSupportedMaps(self,sessionType):
        return bsMap.getMapList("race")

    @classmethod
    def getSettings(self,sessionType):
        return [("Laps",{'minValue':1,"default":3,"increment":1}),
                ("Time Limit",{'choices':[('None',0),('1 Minute',60),
                                          ('2 Minutes',120),('5 Minutes',300),
                                          ('10 Minutes',600),('20 Minutes',1200)],'default':0}),
                ("Mine Spawning",{'choices':[('No Mines',0),('8 Seconds',8000),
                                             ('4 Seconds',4000),('2 Seconds',2000)],'default':4000}),
                ("Epic Mode",{'default':False})]
    
    def __init__(self,session,level,settings):
        self._raceStarted = False
        bsTeamGame.TeamGameActivity.__init__(self,session,level,settings)
        self._scoreBoard = bsScoreBoard.ScoreBoard(self)
        if self.settings['Epic Mode']: self._isSlowMotion = True
        self._scoreSound = bs.getSound("score")
        self._swipSound = bs.getSound("swip")

    def getInstanceDescription(self):
        if self.getSession().getSessionType() == 'teams': tStr = ' Your entire team has to finish.'
        else: tStr = ''

        if self.settings['Laps'] > 1: s = ('Run ${ARG1} laps.'+tStr,self.settings['Laps'])
        else: s = 'Run 1 lap.'+tStr
        return s

    def getInstanceScoreBoardDescription(self):
        if self.settings['Laps'] > 1:
            s = ('run ${ARG1} laps',self.settings['Laps'])
        else:
            s = 'run 1 lap'
        return s
        # return 'run '+str(self.settings['Laps'])+' lap'+('s' if self.settings['Laps'] > 1 else '')

    def onTransitionIn(self):
        bsTeamGame.TeamGameActivity.onTransitionIn(self, music='Epic Race' if self.settings['Epic Mode'] else 'Race')

        self._nubTex = bs.getTexture('nub')
        self._beep1Sound = bs.getSound('raceBeep1')
        self._beep2Sound = bs.getSound('raceBeep2')

        pts = self._level.getDefPoints('racePoint')

        m = self._raceRegionMaterial = bs.Material('Race Point Region')
        m.addAction(conditions=("dstIsMaterial",bs.playerMaterial),
                    actions=(("partMod","collide",True),
                             ("partMod","physical",False),
                             ("call","atConnect",bs.WeakCall(self._handleRacePointCollide))))

        self._regions = []
        for pt in pts:
            self._regions.append(RaceRegion(self,pt,len(self._regions)))

    def _flashPlayer(self,player,scale):
        pos = bs.getAttr(player.actor.node,'translate')
        light = bs.newNode("light",
                           owner=self,
                           attrs={"translate":pos,
                                  "color":(1,1,0),
                                  "heightAttenuated":False,
                                  "radius":0.4})
        bs.gameTimer(500,bs.WeakCall(bs.deleteNode,light))
        bsUtils.animate(light,'intensity',{0:0,100:1.0*scale,500:0})
        
    def _handleRacePointCollide(self):

        regionNode,playerNode = bs.getCollideValues('sourceNode','opposingNode')
        try: player = bs.getNodeDelegate(playerNode).getPlayer()
        except Exception: player = None
        region = bs.getNodeDelegate(regionNode)
        if player is None or region is None: return

        lastRegion = player.data['lastRegion']
        thisRegion = region._index

        if lastRegion != thisRegion:
            # if a player tries to skip regions, smite them
            # ..allow a one region leeway though (its plausable players can get blown over a region, etc)
            if thisRegion > lastRegion + 2:
                if player.isAlive():
                    player.actor.handleMessage(bs.DieMessage())
                    bs.screenMessage("Killing "+player.getName()+" for skipping part of the track!",color=(1,0,0))
            else:
                player.data['lastRegion'] = thisRegion
                if lastRegion >= len(self._regions)-2 and thisRegion == 0:
                    team = player.getTeam()
                    player.data['lap'] = min(self.settings['Laps'],player.data['lap'] + 1)

                    # team lap value is the min of all team players
                    # (entire team has to complete the race)
                    team.data['lap'] = min([p.data['lap'] for p in team.players.values()])

                    # a player is finishing
                    if player.data['lap'] == self.settings['Laps']:

                        # in teams mode, hand out points based on the order players come in
                        if self.getSession().getSessionType() == 'teams':
                            if self._teamFinishPts > 0:
                                self.scoreSet.playerScored(player,self._teamFinishPts,screenMessage=False)
                            self._teamFinishPts -= 25

                        # flash where the player is
                        self._flashPlayer(player,1.0)
                        player.data['finished'] = True
                        player.actor.handleMessage(bs.DieMessage(immediate=True))
                        # makes sure noone behind them passes them in rank while finishing..
                        player.data['distance'] = 9999.0

                        # if the whole team has finished the race..
                        if team.data['lap'] == self.settings['Laps']:
                            bs.playSound(self._scoreSound)
                            player.getTeam().data['finished'] = True
                            player.getTeam().data['time'] = bs.getGameTime()-self._startTime
                            self._checkEndGame()

                        # team has yet to finish..
                        else:
                            bs.playSound(self._swipSound)

                    # they've just finished a lap but not the race..
                    else:
                        bs.playSound(self._swipSound)
                        self._flashPlayer(player,0.3)

                        # print their lap number over their head..
                        try:
                            m = bs.newNode('math',owner=player.actor.node,attrs={'input1':(0,1.9,0),'operation':'add'})
                            bs.connectAttr(player.actor.node,'torsoTranslate',m,'input2')
                            t = bs.newNode('text',
                                           owner=m,
                                           attrs={'text':'Lap '+str(player.data['lap']+1)+'/'+str(self.settings['Laps']),'inWorld':True,'color':(1,1,0,1),
                                                  'scale':[0.015],'hAlign':'center'},
                                           name='spaz lap txt')
                            bs.connectAttr(m,'output',t,'translate')
                            bsUtils.animate(t,'scale',{0:0,200:0.019,2000:0.019,2200:0})
                            bs.gameTimer(2300,bs.WeakCall(bs.deleteNode,m,quiet=True))
                        except Exception,e:
                            print 'Exception printing lap:',e
                        
    def onTeamJoined(self,team):
        team.data['time'] = None
        team.data['lap'] = 0
        team.data['finished'] = False
        self._updateScoreBoard()

    def onPlayerJoined(self,player):
        player.data['lastRegion'] = 0
        player.data['lap'] = 0
        player.data['distance'] = 0.0
        player.data['finished'] = False
        bsTeamGame.TeamGameActivity.onPlayerJoined(self,player)

    def onPlayerLeft(self,player):
        bsTeamGame.TeamGameActivity.onPlayerLeft(self,player)

        # a player leaving disqualifies the team
        # (otherwise in teams mode everyone could just leave except the leading player to win)
        if self.getSession().getSessionType() == 'teams':
            bs.screenMessage(player.getTeam().name+" is disqualified because "+player.getName()+" left",color=(1,1,0))
            player.getTeam().data['finished'] = True
            player.getTeam().data['time'] = None
            player.getTeam().data['lap'] = 0
            bs.playSound(bs.getSound("boo"))
            for player in player.getTeam().players.values():
                player.data['lap'] = 0
                player.data['finished'] = True
                player.actor.handleMessage(bs.DieMessage())

        # delay by one tick so team/player lists will be updated
        bs.gameTimer(1,bs.WeakCall(self._checkEndGame))

    def _updateScoreBoard(self):
        for team in self.teams.values():
            distances = [player.data['distance'] for player in team.players.values()]
            if len(distances) == 0: teamDist = 0
            else: teamDist = min(distances)
            self._scoreBoard.setTeamValue(team,teamDist,self.settings['Laps'],flash=(teamDist >= float(self.settings['Laps'])),showValue=False)

    def onBegin(self):
        self._startTime = bs.getGameTime()
        bsTeamGame.TeamGameActivity.onBegin(self,nameSound=bs.getSound('announceRace'))
        self.setupStandardTimeLimit(self.settings['Time Limit'])
        self._teamFinishPts = 100

        if self.settings['Mine Spawning'] != 0:
            self._raceMines = [{'point':p,'mine':None} for p in self._level.getDefPoints('raceMine')]
            if len(self._raceMines) > 0:
                self._raceMineTimer = bs.Timer(self.settings['Mine Spawning'],bs.WeakCall(self._updateRaceMine),repeat=True)
    
        self._scoreBoardTimer = bs.Timer(250,bs.WeakCall(self._updateScoreBoard),repeat=True)
        self._playerOrderUpdateTimer = bs.Timer(250,bs.WeakCall(self._updatePlayerOrder),repeat=True)

        if self._isSlowMotion:
            tScale = 0.4
            lightY = 50
        else:
            tScale = 1.0
            lightY = 150
        lStart = int(7100*tScale)
        inc = int(1250*tScale)

        bs.gameTimer(lStart,bs.WeakCall(self._doLight1))
        bs.gameTimer(lStart+inc,bs.WeakCall(self._doLight2))
        bs.gameTimer(lStart+2*inc,bs.WeakCall(self._doLight3))
        bs.gameTimer(lStart+3*inc,bs.WeakCall(self._startRace))

        self._startLights = []
        for i in range(4):
            l = bs.newNode('image',
                           owner=self,
                           attrs={'texture':bs.getTexture('nub'),
                                  'opacity':1.0,
                                  'absoluteScale':True,
                                  'translate':(-75+i*50,lightY),
                                  'scale':(50,50),
                                  'attach':'center'})
            bsUtils.animate(l,'opacity',{4000*tScale:0,5000*tScale:1.0,12000*tScale:1.0,12500*tScale:0.0})
            bs.gameTimer(int(13000*tScale),bs.WeakCall(bs.deleteNode,l,quiet=True))
            self._startLights.append(l)

        bs.setAttr(self._startLights[0],'color',(0.2,0,0))
        bs.setAttr(self._startLights[1],'color',(0.2,0,0))
        bs.setAttr(self._startLights[2],'color',(0.2,0.05,0))
        bs.setAttr(self._startLights[3],'color',(0.0,0.3,0))

    def _doLight1(self):
        bs.setAttr(self._startLights[0],'color',(1.0,0,0))
        bs.playSound(self._beep1Sound)
    def _doLight2(self):
        bs.setAttr(self._startLights[1],'color',(1.0,0,0))
        bs.playSound(self._beep1Sound)
    def _doLight3(self):
        bs.setAttr(self._startLights[2],'color',(1.0,0.3,0))
        bs.playSound(self._beep1Sound)
    def _startRace(self):
        bs.setAttr(self._startLights[3],'color',(0.0,1.0,0))
        bs.playSound(self._beep2Sound)
        for player in self.players.values():
            if player.actor is not None:
                try:player.actor.connectToPlayer(player.playerID)
                except Exception,e: print 'Exception in race player connects:',e
        self._raceStarted = True

    def _updatePlayerOrder(self):

        # calc all player distances
        for player in self.players.values():
            try: pos = bsVector.Vector(*bs.getAttr(player.actor.node,'translate'))
            except Exception: pos = None
            if pos is not None:
                rIndex = player.data['lastRegion']
                r1 = self._regions[rIndex]
                r1Pt = bsVector.Vector(*r1._pt[:3])
                r2 = self._regions[0] if rIndex == len(self._regions)-1 else self._regions[rIndex+1]
                r2Pt = bsVector.Vector(*r2._pt[:3])
                r1Dist = (pos-r1Pt).length()
                r2Dist = (pos-r2Pt).length()
                amt = 1.0-(r2Dist/(r2Pt-r1Pt).length())
                amt = player.data['lap'] + (rIndex+amt) * (1.0/len(self._regions))
                player.data['distance'] = amt

        # sort players by distance and update their 
        pList = [[player.data['distance'],player] for player in self.players.values()]
        pList.sort(reverse=True)
        for i,p in enumerate(pList):
            try:
                n = p[1].actor.distanceTxt
                if bs.nodeExists(n): bs.setAttr(n,'text',str(i+1) if p[1].isAlive() else '')
            except Exception: pass

    def _makeMine(self,i):
        m = self._raceMines[i]
        m['mine'] = bsBomb.Bomb(self,bs.Actor.UNRETAINED,translate=m['point'][:3],bombType='landMine')
        m['mine'].handleMessage(bsBomb.ArmMessage())

    def _flashMine(self,i):
        m = self._raceMines[i]
        light = bs.newNode("light",
                           owner=self,
                           attrs={'translate':m['point'][:3],
                                  'color':(1,0.2,0.2),
                                  'radius':0.1,
                                  'heightAttenuated':False})
        bsUtils.animate(light,"intensity",{0:0,100:1.0,200:0},loop=True)
        bs.gameTimer(1000,bs.WeakCall(bs.deleteNode,light))
        
    def _updateRaceMine(self):
        for i in range(3):
            mIndex = random.randrange(len(self._raceMines))
            m = self._raceMines[mIndex]
            if m['mine'] is None or not m['mine'].exists(): break
        if m['mine'] is None or not m['mine'].exists():
            self._flashMine(mIndex)
            bs.gameTimer(950,bs.WeakCall(self._makeMine,mIndex))

    def spawnPlayer(self,player):

        # dont allow spawning if this team is done
        if player.getTeam().data['finished']: return

        pt = self._regions[player.data['lastRegion']]._pt
        regionScale = 0.8 # dont use the full region so we're less likely to spawn off a cliff
        xRange = (-0.5,0.5) if pt[3] == 0 else (-regionScale*pt[3],regionScale*pt[3])
        zRange = (-0.5,0.5) if pt[5] == 0 else (-regionScale*pt[5],regionScale*pt[5])
        pt = (pt[0]+random.uniform(*xRange),pt[1],pt[2]+random.uniform(*zRange))
        spaz = self.spawnSpazForPlayer(player,position=pt,angle=90 if not self._raceStarted else None)

        # prevent controlling of characters before the start of the race
        if not self._raceStarted: spaz.disconnectFromPlayer()

        m = bs.newNode('math',owner=spaz.node,attrs={'input1':(0,1.4,0),'operation':'add'})
        bs.connectAttr(spaz.node,'torsoTranslate',m,'input2')
        spaz.distanceTxt = bs.newNode('text',
                                      owner=spaz.node,
                                      attrs={'text':'',
                                             'inWorld':True,
                                             'color':(1,1,0.4),
                                             'scale':[0.02],
                                             'hAlign':'center'},
                                      name='player distance txt')
        bs.connectAttr(m,'output',spaz.distanceTxt,'translate')

    def _checkEndGame(self):


        # if there's no teams left racing, finish
        teamsStillIn = len([t for t in self.teams.values() if not t.data['finished']])
        if teamsStillIn == 0:
            self._endGame()
            return

        # count the number of teams that have completed the race
        teamsCompleted = len([t for t in self.teams.values() if t.data['finished'] == True and t.data['time'] is not None])

        if teamsCompleted > 0:
            # in teams mode its over as soon as any team finishes the race
            if self.getSession().getSessionType() == 'teams':
                self._endGame()
            else:
                # in ffa we keep the race going while there's still any points to be handed out.
                # find out how many points we have to award and how many teams have finished,
                # and once that matches we're done
                pointsToAward = len(self.getSession().getFFAPointAwards())
                if teamsCompleted >= pointsToAward-teamsCompleted:
                    self._endGame()
                    return
        
    def _endGame(self):
        results = bsTeamGame.Results(scoreName='Time',lowerIsBetter=True,scoreType='milliseconds')
        for t in self.teams.values(): results.setTeamScore(t,t.data['time'])
        # we don't announce a winner in ffa mode since its probably been a while since the first place guy
        # crossed the finish line so it seems odd to be announcing that now..
        self.end(results=results,announceWinningTeam=True if self.getSession().getSessionType() == 'teams' else False)

    def handleMessage(self,m):
        if isinstance(m,bsSpaz.PlayerSpazDeathMessage):
            bsTeamGame.TeamGameActivity.handleMessage(self,m) # augment default
            try:
                player = m.spaz.getPlayer()
                team = player.getTeam()
            except Exception: return
            if not player.data['finished']: self.respawnPlayer(player,respawnTime=1000)
        else: bsTeamGame.TeamGameActivity.handleMessage(self,m)

